// Copyright 2022 MOSEC Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

use crate::tasks::TaskCode;

#[derive(Debug, derive_more::From, derive_more::Display, derive_more::Error)]
pub(crate) enum ServiceError {
    #[display("inference timeout")]
    Timeout,

    #[display("too many request: task queue is full")]
    TooManyRequests,

    #[display("mosec unknown error")]
    UnknownError,

    #[display("SSE inference error: {_0}")]
    SSEError(TaskCode),
}
